/***************************************************************************/
/*  Module:      $Id: compat.c,v 1.1 1995/05/24 00:44:07 maf Exp maf $
/*  Description: porting compatibility routines 
/*  Author:      maf
/* Copyright (c) 1995 Mark Fullmer and The Ohio State University
/***************************************************************************/

#include <time.h> /* needed for the time_t type -ByB */
#include "sendpage.h"

#ifdef NEED_BZERO
/* from perl */
bzero(loc,len)
register char *loc;
register len;
{
	char *retval = loc;

	while (len--)
	*loc++ = 0;
	return retval;
}
#endif /* NEED_BZERO */


#ifdef NEED_BCOPY
/* from perl */
char *
bcopy(from,to,len)
register char *from;
register char *to;
register int len;
{
	char *retval = to;

	if (from - to >= 0) {
	while (len--)
		*to++ = *from++;
	}
	else {
	to += len;
	from += len;
	while (len--)
		*(--to) = *(--from);
	}
	return retval;
}
#endif /* NEED_BCOPY */

